/* /////////////////////////////////////////////////////////////////////////////
 * File:        comstl_memory_functions.h (originally MOMemory.h/.c, ::SynesisCom)
 *
 * Purpose:     COM memory functions.
 *
 * Created:     2nd March 1996
 * Updated:     30th May 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2004, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/comstl
 *                          http://www.comstl.org/
 *
 *              email:      submissions@comstl.org  for submissions
 *                          admin@comstl.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/// \file comstl_memory_functions.h
///
/// COM memory functions.

#ifndef COMSTL_INCL_H_COMSTL_MEMORY_FUNCTIONS
#define COMSTL_INCL_H_COMSTL_MEMORY_FUNCTIONS

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define COMSTL_VER_H_COMSTL_MEMORY_FUNCTIONS_MAJOR     1
# define COMSTL_VER_H_COMSTL_MEMORY_FUNCTIONS_MINOR     4
# define COMSTL_VER_H_COMSTL_MEMORY_FUNCTIONS_REVISION  3
# define COMSTL_VER_H_COMSTL_MEMORY_FUNCTIONS_EDIT      22
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* ////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef COMSTL_INCL_H_COMSTL
# include "comstl.h"                // Include the COMSTL root header
#endif /* !COMSTL_INCL_H_COMSTL */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 *
 * The COMSTL components are contained within the comstl namespace. This is
 * actually an alias for stlsoft::comstl_project,
 *
 * The definition matrix is as follows:
 *
 * _STLSOFT_NO_NAMESPACE    _COMSTL_NO_NAMESPACE    comstl definition
 * ---------------------    --------------------    -----------------
 *  not defined              not defined             = stlsoft::comstl_project
 *  not defined              defined                 not defined
 *  defined                  not defined             comstl
 *  defined                  defined                 not defined
 *
 */

#ifndef _COMSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
/* There is no stlsoft namespace, so must define ::comstl */
namespace comstl
{
# else
/* Define stlsoft::comstl_project */

namespace stlsoft
{

namespace comstl_project
{

# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_COMSTL_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

/// \defgroup comstl_memory_functions COM Memory Functions
/// \ingroup COMSTL
/// \brief These functions expand on the COM API's <code>CoTaskMem*()</code> suite of functions
/// @{

/* /////////////////////////////////////////////////////////////////////////////
 * Functions
 */

/* CoTaskMemGetSize, CoTaskMemDidAlloc & CoTaskMemHeapMinimise
 */

/// \brief Gives the size of a memory block
///
/// This function returns the size of a memory block relative to the COM task
/// alloctor, as per IMalloc::GetSize()
///
/// \param pv Pointer to the memory block
/// \return The size of the memory block (in bytes)
inline ULONG CoTaskMemGetSize(void *pv)
{
    LPMALLOC    lpmalloc;
    ULONG       ulRet;
    HRESULT     hr  =   ::CoGetMalloc(MEMCTX_TASK, &lpmalloc);

    if(SUCCEEDED(hr))
    {
        ulRet = lpmalloc->GetSize(pv);
        lpmalloc->Release();
    }
    else
    {
        ulRet = 0;
    }

    return ulRet;
}

/// \brief Determines allocation ownership of a memory block
///
/// This function returns a value indicating whether a memory block was allocated
/// by the COM task allocator, as per IMalloc::DidAlloc()
///
/// \param pv Pointer to the memory block
/// \return Result indicating ownership
/// \retval 1 The memory block was allocated by the task allocator
/// \retval 0 The memory block was <i>not</i> allocated by the task allocator
/// \retval -1 CoTaskMemDidAlloc() cannot determine whether the memory block was allocated by the task allocator
inline cs_sint_t CoTaskMemDidAlloc(void *pv)
{
    LPMALLOC    lpmalloc;
    cs_sint_t   iRet;
    HRESULT     hr  =   ::CoGetMalloc(MEMCTX_TASK, &lpmalloc);

    if(SUCCEEDED(hr))
    {
        iRet = lpmalloc->DidAlloc(pv);
        lpmalloc->Release();
    }
    else
    {
        iRet = -1;
    }

    return iRet;
}

/// \brief Minimises the heap
///
/// This function minimises the heap as much as possible by releasing unused
/// memory to the operating system, coalescing adjacent free blocks and
/// committing free pages, as as per IMalloc::HeapMinimize().
inline void CoTaskMemHeapMinimise()
{
    LPMALLOC    lpmalloc;

    if(SUCCEEDED(::CoGetMalloc(MEMCTX_TASK, &lpmalloc)))
    {
        lpmalloc->HeapMinimize();
        lpmalloc->Release();
    }
}

/// \brief Minimises the heap
///
/// This function minimises the heap as much as possible by releasing unused
/// memory to the operating system, coalescing adjacent free blocks and
/// committing free pages, as as per IMalloc::HeapMinimize().
inline void CoTaskMemHeapMinimize()
{
    CoTaskMemHeapMinimise();
}

////////////////////////////////////////////////////////////////////////////////
// Unit-testing

#ifdef STLSOFT_UNITTEST

namespace unittest
{
    ss_bool_t test_comstl_memory_functions(unittest_reporter *r)
    {
        using stlsoft::unittest::unittest_initialiser;

        ss_bool_t               bSuccess    =   true;

        unittest_initialiser    init(r, "COMSTL", "memory_functions", __FILE__);

        void    *pv =   ::CoTaskMemAlloc(100);

        if(CoTaskMemGetSize(pv) < 100)
        {
            r->report("Memory block has invalid size ", __LINE__);
            bSuccess = false;
        }

        if(0 == CoTaskMemDidAlloc(pv))
        {
            r->report("Memory block was not recognised by task allocator ", __LINE__);
            bSuccess = false;
        }

        return bSuccess;
    }

    unittest_registrar    unittest_comstl_memory_functions(test_comstl_memory_functions);

} // namespace unittest

#endif /* STLSOFT_UNITTEST */

/* ////////////////////////////////////////////////////////////////////////// */

/// @} // end of group comstl_memory_functions

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _COMSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
} // namespace comstl
# else
} // namespace stlsoft::comstl_project
} // namespace stlsoft
# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_COMSTL_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !COMSTL_INCL_H_COMSTL_MEMORY_FUNCTIONS */

/* ////////////////////////////////////////////////////////////////////////// */
