/* /////////////////////////////////////////////////////////////////////////////
 * File:        stlsoft_array_policies.h
 *
 * Purpose:     Contains the construction policies for the array (fixed and
 *              frame) classes.
 *
 * Created:     1st September 2002
 * Updated:     30th May 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2004, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/stlsoft
 *                          http://www.stlsoft.org/
 *
 *              email:      submissions@stlsoft.org  for submissions
 *                          admin@stlsoft.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/// \file stlsoft_array_policies.h
///
/// Contains the construction policies for the array (fixed and frame) classes.

#ifndef STLSOFT_INCL_H_STLSOFT_ARRAY_POLICIES
#define STLSOFT_INCL_H_STLSOFT_ARRAY_POLICIES

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define STLSOFT_VER_H_STLSOFT_ARRAY_POLICIES_MAJOR     1
# define STLSOFT_VER_H_STLSOFT_ARRAY_POLICIES_MINOR     6
# define STLSOFT_VER_H_STLSOFT_ARRAY_POLICIES_REVISION  2
# define STLSOFT_VER_H_STLSOFT_ARRAY_POLICIES_EDIT      105
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef STLSOFT_INCL_H_STLSOFT
# include "stlsoft.h"   // Include the STLSoft root header
#endif /* !STLSOFT_INCL_H_STLSOFT */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _STLSOFT_NO_NAMESPACE
namespace stlsoft
{
#endif /* _STLSOFT_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Construction policy classes
 */

/// Traits class that controls whether elements in STLSoft arrays are in-place constructed and destroyed
///
/// Defaults to true, indicating that construction and destruction will be
/// performed, but is false for all supported integral and boolean types
///
/// \param T The type
template <ss_typename_param_k T>
struct do_construction
{
    enum
    {
        value = true //!< If \c true the array elements are constructed, otherwise they are not
    };
};

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION

STLSOFT_TEMPLATE_SPECIALISATION
struct do_construction<ss_sint8_t>  { enum { value = false }; };

STLSOFT_TEMPLATE_SPECIALISATION
struct do_construction<ss_uint8_t>  { enum { value = false }; };

STLSOFT_TEMPLATE_SPECIALISATION
struct do_construction<ss_sint16_t> { enum { value = false }; };

STLSOFT_TEMPLATE_SPECIALISATION
struct do_construction<ss_uint16_t> { enum { value = false }; };

STLSOFT_TEMPLATE_SPECIALISATION
struct do_construction<ss_sint32_t> { enum { value = false }; };

STLSOFT_TEMPLATE_SPECIALISATION
struct do_construction<ss_uint32_t> { enum { value = false }; };

#ifdef STLSOFT_CF_64BIT_INT_SUPPORT
STLSOFT_TEMPLATE_SPECIALISATION
struct do_construction<ss_sint64_t> { enum { value = false }; };

STLSOFT_TEMPLATE_SPECIALISATION
struct do_construction<ss_uint64_t> { enum { value = false }; };
#endif /* STLSOFT_CF_64BIT_INT_SUPPORT */

STLSOFT_TEMPLATE_SPECIALISATION
struct do_construction<float>       { enum { value = false }; };

STLSOFT_TEMPLATE_SPECIALISATION
struct do_construction<double>      { enum { value = false }; };

STLSOFT_TEMPLATE_SPECIALISATION
struct do_construction<long double> { enum { value = false }; };

#ifdef __STLSOFT_CF_NATIVE_BOOL_SUPPORT
STLSOFT_TEMPLATE_SPECIALISATION
struct do_construction<ss_bool_t>   { enum { value = false }; };
#endif /* __STLSOFT_CF_NATIVE_BOOL_SUPPORT */

#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */


/// Stipulates that array elements are always constructed
struct do_construction_always
{
    enum { value = true };
};

/// Stipulates that array elements are never constructed
struct do_construction_never
{
    enum { value = false };
};

/*
template<   ss_typename_param_k T
        ,   ss_typename_param_k P
struct inplace_range_construct
{
public:
    void
}
 */

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION

template<   ss_typename_param_k T
        ,   ss_typename_param_k A
        ,   ss_typename_param_k P
        >
struct array_range_initialiser
{
    typedef T   value_type;
    typedef T   *pointer;
    typedef A   allocator_type;
    typedef P   initialisation_policy_type;

public:
    static void construct(allocator_type &ator, pointer p, ss_size_t n)
    {
        if(initialisation_policy_type::value)
        {
            construct(ator, p, n, value_type());
        }
        else
        {
#if 1
            memset(p, 0, n * sizeof(value_type));
#else /* ? 0 */
            fill_n(p, n, 0);
#endif /* 0 */
        }
    }

    static void construct(allocator_type &ator, pointer p, ss_size_t n, value_type const &value)
    {
        for(pointer e = p + n; p != e; ++p)
        {
            ator.construct(p, value);
        }
    }

    static void destroy(allocator_type &ator, pointer p, ss_size_t n)
    {
        if(initialisation_policy_type::value)
        {
            for(pointer e = p + n; p != e; ++p)
            {
                ator.destroy(p);
            }
        }
    }
};

#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _STLSOFT_NO_NAMESPACE
} // namespace stlsoft
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !STLSOFT_INCL_H_STLSOFT_ARRAY_POLICIES */

/* ////////////////////////////////////////////////////////////////////////// */
