/* /////////////////////////////////////////////////////////////////////////////
 * File:        stlsoft_operator_bool.h
 *
 * Purpose:     A robust and portable operator bool generator class.
 *
 * Created:     5th November 2003
 * Updated:     10th June 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2004, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/stlsoft
 *                          http://www.stlsoft.org/
 *
 *              email:      submissions@stlsoft.org  for submissions
 *                          admin@stlsoft.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/// \file stlsoft_operator_bool.h
///
/// A robust and portable operator bool generator class.

#ifndef STLSOFT_INCL_H_STLSOFT_OPERATOR_BOOL
#define STLSOFT_INCL_H_STLSOFT_OPERATOR_BOOL

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define STLSOFT_VER_H_STLSOFT_OPERATOR_BOOL_MAJOR      1
# define STLSOFT_VER_H_STLSOFT_OPERATOR_BOOL_MINOR      3
# define STLSOFT_VER_H_STLSOFT_OPERATOR_BOOL_REVISION   2
# define STLSOFT_VER_H_STLSOFT_OPERATOR_BOOL_EDIT       17
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef STLSOFT_INCL_H_STLSOFT
# include "stlsoft.h"   // Include the STLSoft root header
#endif /* !STLSOFT_INCL_H_STLSOFT */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _STLSOFT_NO_NAMESPACE
namespace stlsoft
{
#endif /* _STLSOFT_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Macros
 */

/// \def STLSOFT_DEFINE_OPERATOR_BOOL_TYPES Defines the types for a type-safe boolean operator
///
/// \param U The Unique type. This is usually the \c class_type member type of the
/// class for which we are providing the boolean operator, but it could be any type.
/// \param G The Generator type. This is the type you use to qualify your call
/// to translate() inside the boolean operator
/// \param B The Boolean type. This is the type of the boolean operator
/// \param

# define STLSOFT_DEFINE_OPERATOR_BOOL_TYPES(U, G, B)                                                \
                                                                                                    \
    typedef stlsoft_ns_qual(operator_bool_generator)<U>::class_type   G;                            \
    typedef G::return_type                                            B

/// \def STLSOFT_DEFINE_OPERATOR_BOOL_TYPES Defines the types for a type-safe boolean operator, for use in templates
///
/// \param U The Unique type. This is usually the \c class_type member type of the
/// class for which we are providing the boolean operator, but it could be any type.
/// \param G The Generator type. This is the type you use to qualify your call
/// to translate() inside the boolean operator
/// \param B The Boolean type. This is the type of the boolean operator
/// \param
///
/// \note This is identical in form and purpose to STLSOFT_DEFINE_OPERATOR_BOOL_TYPES(),
/// except that it must be used for implementing operator boolean for templates

#if defined(__STLSOFT_COMPILER_IS_BORLAND) || \
    (   defined(__STLSOFT_COMPILER_IS_MSVC) && \
        _MSC_VER < 1300)

# define STLSOFT_DEFINE_OPERATOR_BOOL_TYPES_T(U, G, B)                                              \
                                                                                                    \
    STLSOFT_DEFINE_OPERATOR_BOOL_TYPES(U, G, B)

#else /* ? compiler */

# define STLSOFT_DEFINE_OPERATOR_BOOL_TYPES_T(U, G, B)                                              \
                                                                                                    \
    typedef ss_typename_param_k stlsoft_ns_qual(operator_bool_generator)<U>::class_type G;          \
    typedef ss_typename_param_k G::return_type                                          B

#endif /* compiler */

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

/// Template which provides the types and conversion operations for safe and
/// highly-portable "<code>operator bool() const</code>".
template <ss_typename_param_k T>
struct operator_bool_generator
{
public:
    typedef operator_bool_generator<T>  class_type;

#ifdef STLSOFT_CF_OPERATOR_BOOL_AS_OPERATOR_POINTER_TO_MEMBER_SUPPORT
    typedef int class_type::*return_type;

    /// Returns the value representing the true condition
    static return_type true_value()
    {
        return &class_type::i;
    }

private:
    int i;
public:
#else
    typedef class_type const  *return_type;

    /// Returns the value representing the true condition
    static return_type true_value()
    {
        class_type  t;
        void        *p = static_cast<void*>(&t);

        return static_cast<return_type>(p);
    }
#endif // STLSOFT_CF_OPERATOR_BOOL_AS_OPERATOR_POINTER_TO_MEMBER_SUPPORT

    /// Returns the value representing the false condition
    static return_type false_value()
    {
        return static_cast<return_type>(0);
    }

    /// Does the ternary operator for you
#ifdef __STLSOFT_CF_MEMBER_TEMPLATE_FUNCTION_SUPPORT
    template <ss_typename_param_k U>
    static return_type translate(U b)
#else /* ? __STLSOFT_CF_MEMBER_TEMPLATE_FUNCTION_SUPPORT */
    static return_type translate(ss_bool_t b)
#endif // __STLSOFT_CF_MEMBER_TEMPLATE_FUNCTION_SUPPORT
    {
        return b ? true_value() : false_value();
    }

private:
    void operator delete(void*);
};

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _STLSOFT_NO_NAMESPACE
} // namespace stlsoft
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !STLSOFT_INCL_H_STLSOFT_OPERATOR_BOOL */

/* ////////////////////////////////////////////////////////////////////////// */
