/* /////////////////////////////////////////////////////////////////////////////
 * File:        stlsoft_proxy_ptr.h (originally MLBrwPtr.h, ::SynesisStd)
 *
 * Purpose:     Contains the proxy_ptr template class.
 *
 * Created:     17th January 1999
 * Updated:     30th May 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2004, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/stlsoft
 *                          http://www.stlsoft.org/
 *
 *              email:      submissions@stlsoft.org  for submissions
 *                          admin@stlsoft.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/// \file stlsoft_proxy_ptr.h
///
/// Contains the proxy_ptr template class.

#ifndef STLSOFT_INCL_H_STLSOFT_PROXY_PTR
#define STLSOFT_INCL_H_STLSOFT_PROXY_PTR

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define STLSOFT_VER_H_STLSOFT_PROXY_PTR_MAJOR      1
# define STLSOFT_VER_H_STLSOFT_PROXY_PTR_MINOR      4
# define STLSOFT_VER_H_STLSOFT_PROXY_PTR_REVISION   2
# define STLSOFT_VER_H_STLSOFT_PROXY_PTR_EDIT       46
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef STLSOFT_INCL_H_STLSOFT
# include "stlsoft.h"   // Include the STLSoft root header
#endif /* !STLSOFT_INCL_H_STLSOFT */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _STLSOFT_NO_NAMESPACE
namespace stlsoft
{
#endif /* _STLSOFT_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Forward declarations
 */

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION

template <ss_typename_param_k T>
class proxy_ptr;

#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

// class proxy_ptr
/// This class emulates a pointer in all respects, and is simply an aid to
/// self-documentation.
///
/// \param T The value type
template <ss_typename_param_k T>
class proxy_ptr
{
public:
    /// The value type
    typedef T                       value_type;
    /// The current parameterisation of the type
    typedef proxy_ptr<T>            class_type;

// Construction
public:
    /// Construct from a pointer to "borrow"
    ss_explicit_k proxy_ptr(value_type *t)
        : m_value(t)
    {}
    /// Assignment from a new pointer
    proxy_ptr &operator =(value_type *t)
    {
        m_value = t;

        return *this;
    }

// Conversion
public:
    /// Implicit conversion to pointer to the underlying pointer
    operator value_type *()
    {
        return m_value;
    }
    /// Implicit conversion to pointer-to-const to the underlying pointer
    operator const value_type *() const
    {
        return m_value;
    }

    /// Indirection operator
    ///
    /// \note This method does a debug-time assertion that the underlying pointer is non-null
    value_type &operator *() // indirection
    {
        stlsoft_message_assert("Dereferencing a null pointer!", NULL != m_value);

        return *m_value;
    }
    /// Indirection operator
    ///
    /// \note This method does a debug-time assertion that the underlying pointer is non-null
    value_type const &operator *() const // indirection
    {
        stlsoft_message_assert("Dereferencing a null pointer!", NULL != m_value);

        return *m_value;
    }
    /// Member-selection operator
    ///
    /// \note This method does a debug-time assertion that the underlying pointer is non-null
    value_type *operator ->() // member-selection
    {
        stlsoft_message_assert("Dereferencing a null pointer!", NULL != m_value);

        return m_value;
    }
    /// Member-selection operator
    ///
    /// \note This method does a debug-time assertion that the underlying pointer is non-null
    value_type const *operator ->() const // member-selection
    {
        stlsoft_message_assert("Dereferencing a null pointer!", NULL != m_value);

        return m_value;
    }

    /// Returns the underlying pointer value
    value_type *get_ptr()
    {
        return m_value;
    }
    /// Returns the underlying pointer value
    value_type const *get_ptr() const
    {
        return m_value;
    }

    /// Returns the underlying pointer value
    ///
    /// \deprecated
    value_type *GetPointer()
    {
        return m_value;
    }
    /// Returns the underlying pointer value
    ///
    /// \deprecated
    value_type const *GetPointer() const
    {
        return m_value;
    }

    /// Sets the underlying pointer value to null
    void clear()
    {
        m_value = NULL;
    }

// Members
private:
    value_type  *m_value;
};

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _STLSOFT_NO_NAMESPACE
} // namespace stlsoft
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !STLSOFT_INCL_H_STLSOFT_PROXY_PTR */

/* ////////////////////////////////////////////////////////////////////////// */
