/* /////////////////////////////////////////////////////////////////////////////
 * File:        winstl_window_enable_scope.h (originally MWWndScp.h; ::SynesisWin)
 *
 * Purpose:     Window enable-state scoping class.
 *
 * Created:     5th January 1996
 * Updated:     30th May 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2004, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/winstl
 *                          http://www.winstl.org/
 *
 *              email:      submissions@winstl.org  for submissions
 *                          admin@winstl.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/// \file winstl_window_enable_scope.h
///
/// Window enable-state scoping class.

#ifndef WINSTL_INCL_H_WINSTL_WINDOW_ENABLE_SCOPE
#define WINSTL_INCL_H_WINSTL_WINDOW_ENABLE_SCOPE

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define WINSTL_VER_H_WINSTL_WINDOW_ENABLE_SCOPE_MAJOR      1
# define WINSTL_VER_H_WINSTL_WINDOW_ENABLE_SCOPE_MINOR      9
# define WINSTL_VER_H_WINSTL_WINDOW_ENABLE_SCOPE_REVISION   4
# define WINSTL_VER_H_WINSTL_WINDOW_ENABLE_SCOPE_EDIT       65
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef WINSTL_INCL_H_WINSTL
# include "winstl.h"                // Include the WinSTL root header
#endif /* WINSTL_INCL_H_WINSTL */
#ifndef WINSTL_INCL_H_WINSTL_WINDOW_ACCESS
# include "winstl_window_access.h"  // winstl::window_access
#endif /* WINSTL_INCL_H_WINSTL_WINDOW_ACCESS */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _WINSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
/* There is no stlsoft namespace, so must define ::winstl */
namespace winstl
{
# else
/* Define stlsoft::winstl_project */

namespace stlsoft
{

namespace winstl_project
{

# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_WINSTL_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

/// \weakgroup libraries STLSoft Libraries
/// \brief The individual libraries

/// \weakgroup libraries_window Window Library
/// \ingroup libraries
/// \brief This library provides facilities for defining and manipulating GUI windows

/// \defgroup winstl_window_library Window Library (WinSTL)
/// \ingroup WinSTL libraries_window
/// \brief This library provides facilities for defining and manipulating Win32 GUI windows
/// @{

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

// window_enable_scope
/// Provides scoping of the enable status of a window.
///
/// This class provides scoping of the enable status of a window via the API
/// function EnableWindow()..
class window_enable_scope
{
public:
    typedef window_enable_scope class_type;

// Construction
public:
    /// \brief Toggles the window enable state
    ///
    /// Takes a HWND and changes it's current enable-status, which is set back to
    /// the original state in the destructor.
    ///
    /// \param wnd The window whose enable state is to be controlled
    ss_explicit_k window_enable_scope(HWND wnd)
        : m_hwnd(wnd)
        , m_bEnableOnDtor(::IsWindowEnabled(m_hwnd) != false)
    {
        ::EnableWindow(m_hwnd, !m_bEnableOnDtor);
    }

#ifdef __STLSOFT_CF_MEMBER_TEMPLATE_CTOR_SUPPORT
    /// \brief Toggles the window enable state
    ///
    /// Takes a HWND and changes it's current enable-status, which is set back to
    /// the original state in the destructor.
    ///
    /// \param wnd The window whose enable state is to be controlled
    template <ss_typename_param_k W>
    ss_explicit_k window_enable_scope(W &wnd)
        : m_hwnd(get_hwnd(wnd))
        , m_bEnableOnDtor(::IsWindowEnabled(m_hwnd) != false)
    {
        ::EnableWindow(m_hwnd, !m_bEnableOnDtor);
    }
#endif /* __STLSOFT_CF_MEMBER_TEMPLATE_CTOR_SUPPORT */

    /// Modifies the window enable state
    ///
    /// \param wnd The window whose enable state is to be controlled
    /// \param bEnableOnCtor The state to set in the constructor
    /// \param bEnableOnDtor The state it is reset to in the destructor
    window_enable_scope(HWND wnd, ws_bool_t bEnableOnCtor, ws_bool_t bEnableOnDtor)
        : m_hwnd(wnd)
        , m_bEnableOnDtor(bEnableOnDtor)
    {
        ::EnableWindow(m_hwnd, bEnableOnCtor);
    }
#ifdef __STLSOFT_CF_MEMBER_TEMPLATE_CTOR_SUPPORT
    /// Modifies the window enable state
    ///
    /// \param wnd The window whose enable state is to be controlled
    /// \param bEnableOnCtor The state to set in the constructor
    /// \param bEnableOnDtor The state it is reset to in the destructor
    template <ss_typename_param_k W>
    window_enable_scope(W &wnd, ws_bool_t bEnableOnCtor, ws_bool_t bEnableOnDtor)
        : m_hwnd(get_hwnd(wnd))
        , m_bEnableOnDtor(bEnableOnDtor)
    {
        ::EnableWindow(m_hwnd, bEnableOnCtor);
    }
#endif /* __STLSOFT_CF_MEMBER_TEMPLATE_CTOR_SUPPORT */

    /// Resets the enable status
    ~window_enable_scope() winstl_throw_0()
    {
        winstl_static_assert(stlsoft_raw_offsetof(class_type, m_hwnd) < stlsoft_raw_offsetof(class_type, m_bEnableOnDtor));

        ::EnableWindow(m_hwnd, m_bEnableOnDtor);
    }

// Members
private:
    HWND        m_hwnd;
    ws_bool_t   m_bEnableOnDtor;

// Not to be implemented
private:
    window_enable_scope(window_enable_scope const &rhs);
    window_enable_scope const &operator =(window_enable_scope const &rhs);
};

/* ////////////////////////////////////////////////////////////////////////// */

/// @} // end of group winstl_window_library

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _WINSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
} // namespace winstl
# else
} // namespace winstl_project
} // namespace stlsoft
# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_WINSTL_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* WINSTL_INCL_H_WINSTL_WINDOW_ENABLE_SCOPE */

/* ////////////////////////////////////////////////////////////////////////// */
